#import python modules
import numpy as np
import atomwrite

#import compiled fortran library
import MDlib

import sys
import os

#whether or not to use the 3d visualization
UseVisual = False
if UseVisual:
    #import custom visualization library
    import atomvis


#========== GLOBAL VARIABLES ==========
#distance cutoff for pairwise interactions
Cut = 2.5
#set the random number seed; useful for debugging
np.random.seed(342324)


#NOTE:
#everything below assumes unit atomic masses,
#such that forces = accelerations.


def LineSearch(Pos, Dir, dx, EFracTol,
               Accel = 1.5, MaxInc = 10., MaxIter = 10000):
    """Performs a line search along direction Dir.
Input:
    Pos: starting positions, (N,3) array
    Dir: (N,3) array of gradient direction
    dx: initial step amount
    EFracTol: fractional energy tolerance
    Accel: acceleration factor
    MaxInc: the maximum increase in energy for bracketing
    MaxIter: maximum number of iteration steps
Output:
    PEnergy: value of potential energy at minimum along Dir
    Pos: minimum energy (N,3) position array along Dir
"""
    #identify global variables
    global M, Q, L, Cut
    #start the iteration counter
    Iter = 0

    #find the normalized direction    
    NormDir = np.clip(Dir, -1.e100, 1.e100)
    NormDir = NormDir / np.sqrt(np.sum(NormDir * NormDir))

    #take the first two steps and compute energies    
    Dists = [0., dx]
    PEs = [MDlib.calcenergy(Pos + NormDir * x, q1, Q, M, L, Cut, 3, N) for x in Dists]

    #if the second point is not downhill in energy, back
    #off and take a shorter step until we find one
    while PEs[1] > PEs[0]:
        Iter += 1
        dx = dx * 0.5
        Dists[1] = dx
        PEs[1] = MDlib.calcenergy(Pos + NormDir * dx, q1, Q, M, L, Cut, 3, N)
        
    #find a third point
    Dists = Dists + [2. * dx]
    PEs = PEs + [MDlib.calcenergy(Pos + NormDir * 2. * dx, q1, Q, M, L, Cut, 3, N)]
    
    #keep stepping forward until the third point is higher
    #in energy; then we have bracketed a minimum
    while PEs[2] < PEs[1]:
        Iter += 1
            
        #find a fourth point and evaluate energy
        Dists = Dists + [Dists[-1] + dx]
        PEs = PEs + [MDlib.calcenergy(Pos + NormDir * Dists[-1], q1, Q, M, L, Cut, 3, N)]

        #check if we increased too much in energy; if so, back off
        if (PEs[3] - PEs[0]) > MaxInc * (PEs[0] - PEs[2]):
            PEs = PEs[:3]
            Dists = Dists[:3]
            dx = dx * 0.5
        else:
            #shift all of the points over
            PEs = PEs[-3:]
            Dists = Dists[-3:]
            dx = dx * Accel
            
    #we've bracketed a minimum; now we want to find it to high
    #accuracy
    OldPE3 = 1.e300
    while True:
        Iter += 1
        if Iter > MaxIter:
            print "Warning: maximum number of iterations reached in line search."
            break
            
        #store distances for ease of code-reading
        d0, d1, d2 = Dists
        PE0, PE1, PE2 = PEs

        #use a parobolic approximation to estimate the location
        #of the minimum
        d10 = d0 - d1
        d12 = d2 - d1
        Num = d12*d12*(PE0-PE1) - d10*d10*(PE2-PE1)
        Dem = d12*(PE0-PE1) - d10*(PE2-PE1)
        if Dem == 0:
            #parabolic extrapolation won't work; set new dist = 0 
            d3 = 0
        else:
            #location of parabolic minimum
            d3 = d1 + 0.5 * Num / Dem
            
        #compute the new potential energy
        PE3 = MDlib.calcenergy(Pos + NormDir * d3, q1, Q, M, L, Cut, 3, N)
        
        #sometimes the parabolic approximation can fail;
        #check if d3 is out of range < d0 or > d2 or the new energy is higher
        if d3 < d0 or d3 > d2 or PE3 > PE0 or PE3 > PE1 or PE3 > PE2:
            #instead, just compute the new distance by bisecting two
            #of the existing points along the line search
            if abs(d2 - d1) > abs(d0 - d1):
                d3 = 0.5 * (d2 + d1)
            else:
                d3 = 0.5 * (d0 + d1)
            PE3 = MDlib.calcenergy(Pos + NormDir * d3, q1, Q, M, L, Cut, 3, N)
            
        #decide which three points to keep; we want to keep
        #the three that are closest to the minimum
        if d3 < d1:
            if PE3 < PE1:
                #get rid of point 2
                Dists, PEs = [d0, d3, d1], [PE0, PE3, PE1]
            else:
                #get rid of point 0
                Dists, PEs = [d3, d1, d2], [PE3, PE1, PE2]
        else:
            if PE3 < PE1:
                #get rid of point 0
                Dists, PEs = [d1, d3, d2], [PE1, PE3, PE2]
            else:
                #get rid of point 2
                Dists, PEs = [d0, d1, d3], [PE0, PE1, PE3]
        
        #check how much we've changed
        if abs(OldPE3 - PE3) < EFracTol*abs(PE3):
            #the fractional change is less than the tolerance,
            #so we are done and can exit the loop
            break
        OldPE3 = PE3

    #return the position array at the minimum (point 1)        
    PosMin = Pos + NormDir * Dists[1]
    PEMin = PEs[1]
    
    #if using visualization, update the display
    if UseVisual:
        if atomvis.Initialized:
            #update the positions
            atomvis.Update(PosMin)
        else:
            #initialize the visualization window
            atomvis.Init(PosMin)
        
    return PEMin, PosMin

        
def ConjugateGradient(Pos, dx, EFracTolLS, EFracTolCG):
    """Performs a conjugate gradient search.
Input:
    Pos: starting positions, (N,3) array
    dx: initial step amount
    EFracTolLS: fractional energy tolerance for line search
    EFracTolCG: fractional energy tolerance for conjugate gradient
Output:
    PEnergy: value of potential energy at minimum
    Pos: minimum energy (N,3) position array 
"""
    #identify global variables
    global M, Q, q1, L, Cut, Dim, N
    #initial search direction
    Forces = np.zeros_like(Pos)
    PE, Forces = MDlib.calcenergyforces(Pos, q1, Q, M, L, Cut, Forces, 3, N)
    Dir = Forces
    OldPE = 1.e300
    #iterative line searches
    while abs(PE - OldPE) > EFracTolCG * abs(PE):
        OldPE = PE
        PE, Pos = LineSearch(Pos, Dir, dx, EFracTolLS)
        OldForces = Forces.copy()
        PE, Forces = MDlib.calcenergyforces(Pos, q1, Q, M, L, Cut, Forces, 3, N)
        Gamma = np.sum((Forces - OldForces) * Forces) / np.sum(OldForces * OldForces)
        Dir = Forces + Gamma *  Dir
    return PE, Pos


def RescaleVelocities(Vel, T):
    """Rescales velocities in the system to the target temperature using Berendsen Thermostat
Input:
    Vel: (N,3) array of atomic velocities
    T: target temperature
Output:
    Vel: same as above
"""
    #recenter to zero net momentum (assuming all masses same)
    Vel = Vel - Vel.mean(axis=0)
    #find the total kinetic energy
    KE = 0.5 * np.sum(Vel * Vel)
    Tinst = 2*KE/(3*len(Vel)) #DOF=3N=len(Vel)
    #find velocity scale factor from ratios of kinetic energy
    # coupling constant = 0.4 ps, timestep = 1 fs
    VScale = np.sqrt(1+0.0025*(T/Tinst-1))
    Vel = Vel * VScale
    return Vel  


def InitVelocities(N, T):
    """Returns an initial random velocity set.
Input:
    N: number of atoms
    T: target temperature
Output:
    Vel: (N,3) array of atomic velocities
"""
    Vel = np.random.rand(N, 3)
    Vel = RescaleVelocities(Vel, T)
    return Vel


def InitAccel(Pos):
    """Returns the initial acceleration array.
Input:
    Pos: (N,3) array of atomic positions
Output:
    Accel: (N,3) array of acceleration vectors
"""
    #global variables
    global M, Q, q1, L, Cut, Dim, N
    #get the acceleration from the forces
    Forces = np.zeros_like(Pos)
    PEnergy, Accel = MDlib.calcenergyforces(Pos, q1, Q, M, L, Cut, Forces, 3, N)
    return Accel


def InstTemp(Vel):
    """Returns the instantaneous temperature.
Input:
    Vel: (N,3) array of atomic velocities
Output:
    Tinst: float
"""
    return np.sum(Vel * Vel) / (3. * len(Vel))


def InitPositions(N, L):
	"""Returns an array of initial positions of each atom,
Input:
    N: number of atoms
    L: side length vector array, float
Output:
    Pos: (N,3) array of positions
"""
	Pos = np.zeros((N,3), float)
	# initially starts from a extended structure, at the center of the box
	for i in range(N):
		Pos[i,0]=i
		Pos[i,1]=0
		Pos[i,2]=0
	Pos[:,0] -= np.average(Pos[:,0])
	return Pos


def Equilibration(m, dt, nsteps, RescaleFreq, Pos, Vel, Accel ,D):
	KElist=[]; PElist=[]; i=0
	while i < nsteps:
		Pos, Vel, Accel, KE, PE = MDlib.vvintegrate(Pos, q1, Q, Vel, Accel, m, L, Cut, dt, 3, N)
		KElist.append(KE); PElist.append(PE)
		if i % RescaleFreq == 0 : 
			Vel = RescaleVelocities(Vel,T)
			if UseVisual:
				if atomvis.Initialized:
					atomvis.Update(Pos)
				else:
					atomvis.Init(Pos)
		i+=1
	atomwrite.pdbfile("equil"+str(D)+".pdb", Compressed = False).write(Pos)
	return KElist, PElist, Pos, Vel, Accel
	
def Production(m, dt, nsteps, Pos, Vel, Accel, D):
	i =0; k=0; KElist=[]; PElist=[]; distances=[]; Rg=[]; PEplates=[]; 
	# position array for upper and lower plates
	PosPlates = np.zeros([2*int(L[0])*int(L[1]),3])
	for i in range(int(L[0])):
		for j in range(int(L[1])):
			PosPlates[k]=[L[0]/2-i,L[1]/2.-j,L[2]/2.]
			PosPlates[k+int(L[0])*int(L[1])]=[L[0]/2-i,L[1]/2.-j,-L[2]/2.]
			k+=1
	platenames=['P']*int(L[0])*int(L[1])+['N']*int(L[0])*int(L[1])
	Names = ['P']+['A']*10+['N']+platenames
	# save a trajectory
	Pdb = atomwrite.pdbfile("prod"+str(D)+".pdb", AtomNames=Names) # L: box length vector
	while i < nsteps:
		Pos, Vel, Accel, KE, PE = MDlib.vvintegrate(Pos, q1, Q, Vel, Accel, m, L, Cut, dt, 3, N)
		if i % 100 == 0 :
			KElist.append(KE); PElist.append(PE)
			PEplates.append(MDlib.peplates(Pos, M, L, 3, N))
			Pdb.write(np.append(Pos,PosPlates,axis=0)) # write the current position array to the trajectory
			distances.append(np.sqrt(sum((Pos[0]-Pos[11])**2)))# distance btw charged atoms
			Rg.append(RadiusOfGyration(Pos))
		i+=1
	Pdb.close()
	return KElist, PElist, Pos, Vel, Accel, distances, Rg, PEplates

def RescaleMomenta(KElist, PElist, Vel, T):
	avgE=np.average(np.array(KElist)+np.array(PElist))
	avgKE=np.average(np.array(KElist))
	newKE=avgE-PElist[-1]
	VScale = np.sqrt(newKE/KElist[-1])
	Vel = Vel * VScale
	return Vel

def RadiusOfGyration(Pos):
	""" Rg, a float with the corresponding radius of gyration"""
	m=np.average(Pos,axis=0)
	NPos=Pos-m
	r2=sum(np.average(NPos**2,axis=0))
	Rg=np.sqrt(r2)	
	return Rg

import time
start=time.time()

#### GLOBAL VARIABLES ####
T=1.3; Dim=3    # T=270K
N = 12; M=12    # one peptide
rho = 0.01
dt=0.001
RescaleFreq=100

q1 = np.zeros(N)
# a positive charge on one end, 
# and a negative charge on the other end
q1[0]=1; q1[11]=-1
Q = np.zeros([N,N]) # q1*q2
Q[0,11]=-1
Q[11,0]=-1
Q[0,0]=1
Q[11,11]=1

##### NVT MD, Berendsen thermostat #######
# D is the distance btw two plates 
Dlist=[2.1, 2.2, 2.3,2.4,2.5,2.7,3.0,3.2,3.5,4.0,5.0,6.0,8.0]

for D in Dlist:
	L = np.array([np.sqrt((N/rho)/D),np.sqrt((N/rho)/D),D],dtype=float)
	Cut = L[0]/2.; print L
	NStepsEquil1=100000; NStepsEquil2=100000; NStepsProd=3000000 
	# minimization
	Pos = InitPositions(N, L)
	PE, Pos = ConjugateGradient(Pos, 0.001, 1.e-8, 1.e-10)
	# initialize MD run
	Accel = InitAccel(Pos)
	Vel = InitVelocities(N,T)
	
	#### equilibration ####
	KElist, PElist, Pos, Vel, Accel = Equilibration(M, dt, NStepsEquil1, RescaleFreq, Pos, Vel, Accel, D)
	KElist, PElist, Pos, Vel, Accel = Equilibration(M, dt, NStepsEquil2, RescaleFreq, Pos, Vel, Accel, D)
	# check PE vs. time
	import matplotlib.pyplot as plt
	fig=plt.figure(figsize=(7,5))
	sp1=plt.subplot(1,1,1)
	sp1.plot(dt*np.arange(len(PElist)),PElist)
	fig.savefig("equil"+str(D)+".png",dpi=300)
	# write KElist and PElist
	s=""
	for i in range(len(KElist)):
		s+= "%-12.4f \t %-12.4f \n" % (KElist[i], PElist[i])
	f=open('equil'+str(D)+'.txt','w')
	f.write(s)
	f.close()
	
	#### production ####
	Vel = RescaleMomenta(KElist, PElist, Vel, T)
	KElist, PElist, Pos, Vel, Accel, distances, Rg, PEplates = Production(M, dt, NStepsProd, Pos, Vel, Accel,D)
	# write results
	s=""
	for i in range(len(KElist)):
		s+= "%-12.6e \t %-12.6e \t %-12.6e \t %-12.6e \t %-12.6e \n" % (KElist[i], PElist[i], PEplates[i], distances[i], Rg[i])
	f=open('results'+str(D)+'.txt','w')
	f.write(s)
	f.close()

	
end=time.time()-start
print "time="+str(end)